"""Constants."""
from datetime import timedelta
from enum import IntEnum

HEADERS = {"content-type": "application/json"}

HEADER_DAILY_API_LIMIT = "X-RateLimit-Limit-hour"

DAILY = "daily"
HOURLY = "hourly"
NOWCAST = "nowcast"
FORECASTS = "forecasts"

TIMESTEP_HOURLY = "1h"
TIMESTEP_DAILY = "1d"

TYPE_WEATHER = "weather"
TYPE_POLLEN = "pollen"
TYPE_AIR_QUALITY = "air_quality"
TYPE_FIRE = "fire"
TYPE_SOLAR = "solar"
TYPE_PRECIPITATION = "precipitation"

# V4 constants
BASE_URL_V4 = "https://api.tomorrow.io/v4/timelines"
CURRENT = "current"

ONE_DAY = timedelta(days=1)
ONE_HOUR = timedelta(hours=1)
THIRTY_MINUTES = timedelta(minutes=30)
FIFTEEN_MINUTES = timedelta(minutes=15)
FIVE_MINUTES = timedelta(minutes=5)
ONE_MINUTE = timedelta(minutes=1)
REALTIME = timedelta(0)

VALID_TIMESTEPS = (
    ONE_DAY,
    ONE_HOUR,
    THIRTY_MINUTES,
    FIFTEEN_MINUTES,
    FIVE_MINUTES,
    ONE_MINUTE,
    REALTIME,
)


class UVDescription(IntEnum):
    """UV index and health concern description."""

    LOW = 0
    MODERATE = 1
    HIGH = 2
    VERY_HIGH = 3
    EXTREME = 4


def convert_uv_value(value: int) -> UVDescription:
    """Convert a UV index or health concern value to a description."""
    if value < 3:
        return UVDescription.LOW
    if value < 6:
        return UVDescription.MODERATE
    if value < 8:
        return UVDescription.HIGH
    if value < 11:
        return UVDescription.VERY_HIGH

    return UVDescription.EXTREME


class MoonPhase(IntEnum):
    """Moon phase."""

    NEW = 0
    WAXING_CRESCENT = 1
    FIRST_QUARTER = 2
    WAXING_GIBBOUS = 3
    FULL = 4
    WANING_GIBBOUS = 5
    THIRD_QUARTER = 6
    WANING_CRESCENT = 7


class FloodIndex(IntEnum):
    """Flood index."""

    MINOR = 1
    MODERATE = 2
    SIGNIFICANT = 3
    MAJOR = 4
    CATASTROPHIC = 5


class PrecipitationType(IntEnum):
    """Precipitation types."""

    NONE = 0
    RAIN = 1
    SNOW = 2
    FREEZING_RAIN = 3
    ICE_PELLETS = 4


class PollenIndex(IntEnum):
    """Pollen index."""

    NONE = 0
    VERY_LOW = 1
    LOW = 2
    MEDIUM = 3
    HIGH = 4
    VERY_HIGH = 5


class PrimaryPollutantType(IntEnum):
    """Primary pollutant type."""

    PM25 = 0
    PM10 = 1
    O3 = 2
    NO2 = 3
    CO = 4
    SO2 = 5


class HealthConcernType(IntEnum):
    """Health concern type."""

    GOOD = 0
    MODERATE = 1
    UNHEALTHY_FOR_SENSITIVE_GROUPS = 2
    UNHEALTHY = 3
    VERY_UNHEALTHY = 4
    HAZARDOUS = 5


class WeatherCode(IntEnum):
    """Weather codes"""

    UNKNOWN = 0
    CLEAR = 1000
    CLOUDY = 1001
    MOSTLY_CLEAR = 1100
    PARTLY_CLOUDY = 1101
    MOSTLY_CLOUDY = 1102
    FOG = 2000
    LIGHT_FOG = 2100
    LIGHT_WIND = 3000
    WIND = 3001
    STRONG_WIND = 3002
    DRIZZLE = 4000
    RAIN = 4001
    LIGHT_RAIN = 4200
    HEAVY_RAIN = 4201
    SNOW = 5000
    FLURRIES = 5001
    LIGHT_SNOW = 5100
    HEAVY_SNOW = 5101
    FREEZING_DRIZZLE = 6000
    FREEZING_RAIN = 6001
    LIGHT_FREEZING_RAIN = 6200
    HEAVY_FREEZING_RAIN = 6201
    ICE_PELLETS = 7000
    HEAVY_ICE_PELLETS = 7101
    LIGHT_ICE_PELLETS = 7102
    THUNDERSTORM = 8000



"""Exceptions for pytomorrowio."""


class TomorrowioException(Exception):
    """Base Exception class for pytomorrowio."""


class MalformedRequestException(TomorrowioException):
    """Raised when request was malformed."""


class InvalidAPIKeyException(TomorrowioException):
    """Raised when API key is invalid."""


class RateLimitedException(TomorrowioException):
    """Raised when API rate limit has been exceeded."""


class UnknownException(TomorrowioException):
    """Raised when unknown error occurs."""


class CantConnectException(TomorrowioException):
    """Raise when client can't connect to Tomorrowio API."""


class InvalidTimestep(TomorrowioException):
    """Raise when an invalid timestep is specified."""



"""Tests for `pytomorrowio` package."""
import json
import re
import sys
from datetime import datetime
from types import SimpleNamespace
from typing import Mapping, Sequence

from aiohttp import (
    ClientSession,
    TraceConfig,
    TraceRequestChunkSentParams,
    TraceRequestEndParams,
)

if sys.version_info < (3, 8):
    from asynctest import patch
else:
    from unittest.mock import patch

from pytomorrowio import TomorrowioV4
from pytomorrowio.const import (
    FIVE_MINUTES,
    ONE_DAY,
    ONE_HOUR,
    TIMESTEP_DAILY,
    TIMESTEP_HOURLY,
    TYPE_POLLEN,
    TYPE_PRECIPITATION,
    TYPE_WEATHER,
)

GPS_COORD = (28.4195, -81.5812)


def create_trace_config() -> TraceConfig:
    async def on_request_chunk_sent(
        session: ClientSession,
        trace_config_ctx: SimpleNamespace,
        params: TraceRequestChunkSentParams,
    ):
        trace_config_ctx.request_body = params.chunk

    async def on_request_end(
        session: ClientSession,
        trace_config_ctx: SimpleNamespace,
        params: TraceRequestEndParams,
    ):
        print("Request:")
        print(params.url)
        for k, v in params.headers.items():
            print(f"  {k}: {v}")
        if trace_config_ctx.request_body:
            print(trace_config_ctx.request_body)
        print()
        print("Response:")
        for k, v in sorted(params.response.headers.items()):
            print(f"  {k}: {v}")
        resp = await params.response.json()
        print(json.dumps(resp, indent=2))

    trace_config = TraceConfig()
    trace_config.on_request_chunk_sent.append(on_request_chunk_sent)
    trace_config.on_request_end.append(on_request_end)
    return trace_config


def create_call_api_mock():
    return patch.object(TomorrowioV4, "_call_api")


def load_json(file_name: str):
    with open(f"tests/fixtures/{file_name}", "r") as file:
        return json.load(file)


async def _test_capture_request_and_response():
    # Remove leading underscore to capture request & response
    async with ClientSession(trace_configs=[create_trace_config()]) as session:
        api = TomorrowioV4("real_api_key", *GPS_COORD, session=session)

        available_fields = api.available_fields(
            ONE_DAY, [TYPE_POLLEN, TYPE_PRECIPITATION, TYPE_WEATHER]
        )
        await api.forecast_daily(available_fields)
        assert False


async def test_timelines_hourly_good():
    with create_call_api_mock() as mock:
        mock.return_value = load_json("timelines_hourly_good.json")

        api = TomorrowioV4("bogus_api_key", *GPS_COORD)
        available_fields = api.available_fields(
            ONE_HOUR, [TYPE_POLLEN, TYPE_PRECIPITATION, TYPE_WEATHER]
        )
        res = await api.forecast_hourly(available_fields)
        mock.assert_called_once()

    assert res is not None
    assert isinstance(res, Mapping)

    data = res.get("data")
    assert isinstance(data, Mapping)

    timelines = data.get("timelines")
    assert isinstance(timelines, Sequence)
    assert len(timelines) == 1

    timeline = timelines[0]
    assert isinstance(timeline, Mapping)

    assert timeline.get("timestep") == TIMESTEP_HOURLY

    intervals = timeline.get("intervals")
    assert isinstance(intervals, Sequence)
    assert len(intervals) == 109

    for interval in intervals:
        assert isinstance(interval, Mapping)

        start_time = interval.get("startTime")
        assert isinstance(start_time, str)

        # Might be fixed in Python 3.11 - https://github.com/python/cpython/issues/80010
        # Maybe pytomorrowio should do the replacements to improve user experience
        start_time = re.sub("Z$", "+00:00", start_time)
        datetime.fromisoformat(start_time)

        values = interval.get("values")
        assert isinstance(values, Mapping)

        assert set(values) == set(available_fields)


async def test_timelines_daily_good():
    with create_call_api_mock() as mock:
        mock.return_value = load_json("timelines_daily_good.json")

        api = TomorrowioV4("bogus_api_key", *GPS_COORD)
        available_fields = api.available_fields(
            ONE_DAY, [TYPE_POLLEN, TYPE_PRECIPITATION, TYPE_WEATHER]
        )
        res = await api.forecast_daily(available_fields)
        mock.assert_called_once()

    assert res is not None
    assert isinstance(res, Mapping)

    data = res.get("data")
    assert isinstance(data, Mapping)

    timelines = data.get("timelines")
    assert isinstance(timelines, Sequence)
    assert len(timelines) == 1

    timeline = timelines[0]
    assert isinstance(timeline, Mapping)

    assert timeline.get("timestep") == TIMESTEP_DAILY

    intervals = timeline.get("intervals")
    assert isinstance(intervals, Sequence)
    assert len(intervals) == 15

    expected_values = api.convert_fields_to_measurements(available_fields)

    for interval in intervals:
        assert isinstance(interval, Mapping)

        start_time = interval.get("startTime")
        assert isinstance(start_time, str)

        values = interval.get("values")
        assert isinstance(values, Mapping)

        assert set(values) == set(expected_values)


async def test_timelines_5min_good():
    with create_call_api_mock() as mock:
        mock.return_value = load_json("timelines_5min_good.json")

        api = TomorrowioV4("bogus_api_key", *GPS_COORD)
        available_fields = api.available_fields(
            FIVE_MINUTES, [TYPE_POLLEN, TYPE_PRECIPITATION, TYPE_WEATHER]
        )
        res = await api.forecast_nowcast(available_fields)

    assert res is not None
    assert isinstance(res, Mapping)

    data = res.get("data")
    assert isinstance(data, Mapping)

    timelines = data.get("timelines")
    assert isinstance(timelines, Sequence)
    assert len(timelines) == 1

    timeline = timelines[0]
    assert isinstance(timeline, Mapping)

    assert timeline.get("timestep") == "5m"



"""Main module."""
import asyncio
import json
import logging
from datetime import datetime, timedelta, timezone
from http import HTTPStatus
from typing import Any, Dict, List, Optional, Union

from aiohttp import ClientConnectionError, ClientSession

from .const import (
    BASE_URL_V4,
    CURRENT,
    DAILY,
    FORECASTS,
    HEADER_DAILY_API_LIMIT,
    HEADERS,
    HOURLY,
    NOWCAST,
    ONE_DAY,
    ONE_HOUR,
    REALTIME,
    TIMESTEP_DAILY,
    TIMESTEP_HOURLY,
    VALID_TIMESTEPS,
)
from .exceptions import (
    CantConnectException,
    InvalidAPIKeyException,
    InvalidTimestep,
    MalformedRequestException,
    RateLimitedException,
    UnknownException,
)
from .fields import FIELDS_V4
from .helpers import async_to_sync

_LOGGER = logging.getLogger(__name__)


def process_v4_fields(
    fields: List[str], timestep: timedelta, write_log: bool = True
) -> List[str]:
    """
    Filter v4 field list to only include valid fields for a given endpoint.

    Optionally logs a warning when fields get filtered out.
    """
    valid_fields = [field for field in fields if field in FIELDS_V4]
    if write_log and len(valid_fields) < len(fields):
        _LOGGER.warning(
            "Removed invalid fields: %s", list(set(fields) - set(valid_fields))
        )
    processed_fields = [
        field for field in valid_fields if timestep <= FIELDS_V4[field].max_timestep
    ]
    if write_log and len(processed_fields) < len(valid_fields):
        _LOGGER.warning(
            "Removed fields not available for `%s` timestep: %s",
            timestep,
            list(set(valid_fields) - set(processed_fields)),
        )
    return processed_fields


def dt_to_utc(input_dt: datetime) -> datetime:
    """If input datetime has a timezone defined, convert to UTC."""
    if input_dt and input_dt.tzinfo:
        return input_dt.astimezone(timezone.utc)
    return input_dt


class TomorrowioV4:
    """Async class to query the Tomorrow.io v4 API."""

    def __init__(
        self,
        apikey: str,
        latitude: Union[int, float, str],
        longitude: Union[int, float, str],
        unit_system: str = "imperial",
        session: Optional[ClientSession] = None,
    ) -> None:
        """Initialize Tomorrow.io API object."""
        if unit_system.lower() not in ("metric", "imperial", "si", "us"):
            raise ValueError("`unit_system` must be `metric` or `imperial`")
        if unit_system.lower() == "si":
            unit_system = "metric"
        elif unit_system.lower() == "us":
            unit_system = "imperial"

        self._apikey = apikey
        self.unit_system = unit_system.lower()
        self._session = session
        self._params = {
            "location": f"{float(latitude)},{float(longitude)}",
            "units": self.unit_system,
        }
        self._headers = {**HEADERS, "apikey": self._apikey}
        self._max_requests_per_day: Optional[int] = None

    @property
    def max_requests_per_day(self) -> Optional[int]:
        """Return the maximum number of requests per day."""
        return self._max_requests_per_day

    @staticmethod
    def convert_fields_to_measurements(fields: List[str]) -> List[str]:
        """Converts general field list into fields with measurements."""
        field_list = []
        for field in fields:
            measurements = FIELDS_V4[field].measurements
            if len(measurements) < 2:
                field_list.append(field)
            else:
                field_list.extend(
                    [f"{field}{measurement}" for measurement in measurements]
                )

        return field_list

    @staticmethod
    def available_fields(
        timestep: timedelta, types: Optional[List[str]] = None
    ) -> List[str]:
        "Return available fields for a given timestep."
        if timestep not in VALID_TIMESTEPS:
            raise InvalidTimestep(f"{timestep} is not a valid 'timestep' parameter")
        fields = process_v4_fields(list(FIELDS_V4), timestep, write_log=False)

        if types:
            return [field for field in fields if FIELDS_V4[field].type in types]

        return fields

    async def _call_api(self, params: Dict[str, Any]) -> Dict[str, Any]:
        """Call Tomorrow.io API."""
        if self._session:
            return await self._make_call(params, self._session)

        async with ClientSession() as session:
            return await self._make_call(params, session)

    async def _make_call(
        self, params: Dict[str, Any], session: ClientSession
    ) -> Dict[str, Any]:
        try:
            resp = await session.post(
                BASE_URL_V4,
                headers=self._headers,
                data=json.dumps({**self._params, **params}),
            )
            resp_json = await resp.json()
        except ClientConnectionError as error:
            raise CantConnectException() from error

        if resp.status == HTTPStatus.OK:
            max_requests = resp.headers[HEADER_DAILY_API_LIMIT]
            if max_requests != self._max_requests_per_day:
                self._max_requests_per_day = int(max_requests)
            return resp_json
        if resp.status == HTTPStatus.BAD_REQUEST:
            raise MalformedRequestException(resp_json, resp.headers)
        if resp.status in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
            raise InvalidAPIKeyException(resp_json, resp.headers)
        if resp.status == HTTPStatus.TOO_MANY_REQUESTS:
            raise RateLimitedException(resp_json, resp.headers)
        raise UnknownException(resp_json, resp.headers)

    async def realtime(self, fields: List[str]) -> Dict[str, Any]:
        """Return realtime weather conditions from Tomorrow.io API."""
        return await self._call_api(
            {
                "fields": process_v4_fields(fields, REALTIME),
                "timesteps": ["current"],
            }
        )

    async def _forecast(
        self,
        timestep: timedelta,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """Return forecast data from Tomorrow.io's API for a given time period."""
        if timestep not in VALID_TIMESTEPS:
            raise InvalidTimestep(f"{timestep} is not a valid 'timestep' parameter")
        fields = process_v4_fields(fields, timestep)
        if timestep > ONE_HOUR:
            fields = self.convert_fields_to_measurements(fields)

        params: Dict[str, Any] = {
            "fields": fields,
            **kwargs,
        }
        if timestep == ONE_DAY:
            params["timesteps"] = [TIMESTEP_DAILY]
        elif timestep == ONE_HOUR:
            params["timesteps"] = [TIMESTEP_HOURLY]
        else:
            params["timesteps"] = [f"{int(timestep.total_seconds()/60)}m"]

        if start_time:
            if not start_time.tzinfo:
                start_time.replace(tzinfo=timezone.utc)
            params["startTime"] = start_time.replace(microsecond=0).isoformat()
        else:
            start_time = datetime.now(tz=timezone.utc)
        if duration:
            end_time = (start_time + duration).replace(microsecond=0)
            params["endTime"] = end_time.isoformat()

        return await self._call_api(params)

    async def forecast_nowcast(
        self,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
        timestep: int = 5,
    ) -> Dict[str, Any]:
        """Return forecast data from Tomorrow.io's NowCast API for a given time period."""
        if timestep not in (1, 5, 15, 30):
            raise InvalidTimestep
        return await self._forecast(
            timedelta(minutes=timestep),
            fields,
            start_time=start_time,
            duration=duration,
        )

    async def forecast_daily(
        self,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
    ) -> Dict[str, Any]:
        """Return daily forecast data from Tomorrow.io's API for a given time period."""
        return await self._forecast(
            ONE_DAY, fields, start_time=start_time, duration=duration
        )

    async def forecast_hourly(
        self,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
    ) -> Dict[str, Any]:
        """Return hourly forecast data from Tomorrow.io's API for a given time period."""
        return await self._forecast(
            ONE_HOUR, fields, start_time=start_time, duration=duration
        )

    async def realtime_and_all_forecasts(
        self,
        realtime_fields: List[str],
        forecast_or_nowcast_fields: List[str],
        hourly_fields: List[str] = None,
        daily_fields: List[str] = None,
        nowcast_timestep: int = 5,
    ) -> Dict[str, Any]:
        """
        Return realtime weather and all forecasts.

        If `hourly_fields` and `daily_fields` are not provided,
        `forecast_or_nowcast_fields` will be used to get nowcast, hourly, and daily
        data.
        """
        ret_data = {}
        data = await self._call_api(
            {
                "timesteps": ["current"],
                "fields": realtime_fields,
            }
        )
        if (
            "data" in data
            and "timelines" in data["data"]
            and "intervals" in data["data"]["timelines"][0]
            and "values" in data["data"]["timelines"][0]["intervals"][0]
        ):
            ret_data[CURRENT] = data["data"]["timelines"][0]["intervals"][0]["values"]

        forecasts = ret_data.setdefault(FORECASTS, {})
        start_time = datetime.now(tz=timezone.utc).isoformat()

        if not hourly_fields and not daily_fields:
            data = await self._call_api(
                {
                    "timesteps": [
                        f"{nowcast_timestep}m",
                        TIMESTEP_HOURLY,
                        TIMESTEP_DAILY,
                    ],
                    "fields": forecast_or_nowcast_fields,
                    "startTime": start_time,
                }
            )
            if "data" in data and "timelines" in data["data"]:
                for timeline in data["data"]["timelines"]:
                    if timeline["timestep"] == TIMESTEP_DAILY:
                        key = DAILY
                    elif timeline["timestep"] == TIMESTEP_HOURLY:
                        key = HOURLY
                    else:
                        key = NOWCAST
                    forecasts[key] = timeline["intervals"]
        else:
            data = await self._call_api(
                {
                    "timesteps": [f"{nowcast_timestep}m"],
                    "fields": forecast_or_nowcast_fields,
                    "startTime": start_time,
                }
            )
            if "data" in data and "timelines" in data["data"]:
                forecasts[NOWCAST] = data["data"]["timelines"][0]["intervals"]

            for field_list, timestep, key in (
                (hourly_fields, TIMESTEP_HOURLY, HOURLY),
                (daily_fields, TIMESTEP_DAILY, DAILY),
            ):
                if field_list:
                    await asyncio.sleep(1)
                    data = await self._call_api(
                        {
                            "timesteps": [timestep],
                            "fields": field_list,
                            "startTime": start_time,
                        }
                    )
                    if "data" in data and "timelines" in data["data"]:
                        forecasts[key] = data["data"]["timelines"][0]["intervals"]
        return ret_data


class TomorrowioV4Sync(TomorrowioV4):
    """Synchronous class to query the Tomorrow.io API."""

    def __init__(
        self,
        apikey: str,
        latitude: Union[int, float, str],
        longitude: Union[int, float, str],
        unit_system: str = "imperial",
    ) -> None:
        """Initialize Synchronous Tomorrow.io v4 API object."""
        super().__init__(apikey, latitude, longitude, unit_system)

    @async_to_sync
    async def realtime(self, fields: List[str]) -> Dict[str, Any]:
        """Return realtime weather conditions from Tomorrow.io API."""
        return await super().realtime(fields)

    @async_to_sync
    async def forecast_nowcast(
        self,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
        timestep: int = 5,
    ) -> Dict[str, Any]:
        """Return forecast data from Tomorrow.io's NowCast API for a given time period."""
        return await super().forecast_nowcast(fields, start_time, duration, timestep)

    @async_to_sync
    async def forecast_daily(
        self,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
    ) -> Dict[str, Any]:
        """Return daily forecast data from Tomorrow.io's API for a given time period."""
        return await super().forecast_daily(fields, start_time, duration)

    @async_to_sync
    async def forecast_hourly(
        self,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
    ) -> Dict[str, Any]:
        """Return hourly forecast data from Tomorrow.io's API for a given time period."""
        return await super().forecast_hourly(fields, start_time, duration)

    @async_to_sync
    async def realtime_and_all_forecasts(
        self,
        realtime_fields: List[str],
        forecast_or_nowcast_fields: List[str],
        hourly_fields: List[str] = None,
        daily_fields: List[str] = None,
        nowcast_timestep: int = 5,
    ) -> Dict[str, Any]:
        """
        Return realtime weather and all forecasts.

        If `hourly_fields` and `daily_fields` are not provided,
        `forecast_or_nowcast_fields` will be used to get nowcast, hourly, and daily
        data.
        """
        return await super().realtime_and_all_forecasts(
            realtime_fields=realtime_fields,
            forecast_or_nowcast_fields=forecast_or_nowcast_fields,
            hourly_fields=hourly_fields,
            daily_fields=daily_fields,
            nowcast_timestep=nowcast_timestep,
        )



"""Top-level package for Python Tomorrow.io API."""

from .pytomorrowio import TomorrowioV4, TomorrowioV4Sync

__author__ = """raman325"""
__email__ = "7243222+raman325@users.noreply.github.com"
__version__ = "0.2.1"
